@@ -1,6 +1,6 @@
 package forks::BerkeleyDB;
 
-$VERSION = 0.053;
+$VERSION = 0.054;
 
 package
 	CORE::GLOBAL;	#hide from PAUSE
@@ -22,21 +22,62 @@ use constant DEBUG => forks::BerkeleyDB::Config::DEBUG();
 use constant ENV_ROOT => forks::BerkeleyDB::Config::ENV_ROOT();
 use constant ENV_SUBPATH => forks::BerkeleyDB::Config::ENV_SUBPATH();
 use constant ENV_PATH => forks::BerkeleyDB::Config::ENV_PATH();
+use constant ENV_PATH_LOCKSIG => forks::BerkeleyDB::Config::ENV_PATH_LOCKSIG();
 
-our $bdb_env;	#berkeleydb environment
-our $bdb_locksig_env;	#berkeleydb lock/signal environment
+my $bdb_env;	#berkeleydb environment
+my $bdb_locksig_env;	#berkeleydb lock/signal environment
 
+### environment variable controls ###
+my $USE_BDB_LOCKS;
+my $BDB_ENV_CHMOD_OCTVAL;
+my $BDB_ENV_CHOWN_ID;
+my $BDB_ENV_CHGRP_ID;
 BEGIN {
+	no warnings 'redefine';
+
 	### allow user to enable BDB locks (disabled by default) ###
 	if (exists $ENV{'THREADS_BDB_LOCKS'}) {	#TODO: convert to import argument in future (i.e. lock_model => 'bdb')
 		$ENV{'THREADS_BDB_LOCKS'} =~ m#^(.*)$#s;
-		no warnings 'redefine';
-		*USE_BDB_LOCKS = $ENV{'THREADS_BDB_LOCKS'} ? sub { 1 } : sub { 0 };
+		$USE_BDB_LOCKS = $ENV{'THREADS_BDB_LOCKS'} ? 1 : 0;
 	} else {
-		*USE_BDB_LOCKS = sub { 0 };
+		$USE_BDB_LOCKS = 0 ;
 	}
+	*USE_BDB_LOCKS = sub { $USE_BDB_LOCKS };
+
+	### allow user to set ENV file permissions; default is 0666 (octal) ###
+	if (exists $ENV{'THREADS_BDB_ENV_CHMOD'}) {	#TODO: convert to import argument in future (i.e. env_chmod => '0666')
+		$ENV{'THREADS_BDB_ENV_CHMOD'} =~ m#^(.*)$#s;
+		$BDB_ENV_CHMOD_OCTVAL = $ENV{'THREADS_BDB_ENV_CHMOD'} =~ m/^0[0-6]{3}/o
+			? oct($ENV{'THREADS_BDB_ENV_CHMOD'})
+			: 0666;
+	} else {
+		$BDB_ENV_CHMOD_OCTVAL = 0666;
+	}
+	*BDB_ENV_CHMOD_OCTVAL = sub { $BDB_ENV_CHMOD_OCTVAL };
+
+	### allow user to set ENV directory structure owner ###
+	if (exists $ENV{'THREADS_BDB_ENV_CHOWN'}) {	#TODO: convert to import argument in future (i.e. env_chown => 'root')
+		$ENV{'THREADS_BDB_ENV_CHOWN'} =~ m#^(.*)$#s;
+		my $uid = (getpwnam($ENV{'THREADS_BDB_ENV_CHOWN'}))[2];
+		$BDB_ENV_CHOWN_ID = defined $uid ? $uid : -1;
+	} else {
+		$BDB_ENV_CHOWN_ID = -1;
+	}
+	*BDB_ENV_CHOWN_ID = sub { $BDB_ENV_CHOWN_ID };
+
+	### allow user to set ENV directory structure group ###
+	if (exists $ENV{'THREADS_BDB_ENV_CHGRP'}) {	#TODO: convert to import argument in future (i.e. env_chgrp => 'sys')
+		$ENV{'THREADS_BDB_ENV_CHGRP'} =~ m#^(.*)$#s;
+		my $gid = (getgrnam($ENV{'THREADS_BDB_ENV_CHGRP'}))[2];
+		$BDB_ENV_CHGRP_ID = defined $gid ? $gid : -1;
+	} else {
+		$BDB_ENV_CHGRP_ID = -1;
+	}
+	*BDB_ENV_CHGRP_ID = sub { $BDB_ENV_CHGRP_ID };
 }
 
+use constant DEFAULT_ENV_PATHS => (ENV_PATH, (USE_BDB_LOCKS() ? ENV_PATH_LOCKSIG : ()));
+
 BEGIN {
 	$forks::DEFER_INIT_BEGIN_REQUIRE = 1;	#feature in forks 0.26 and later
 	require forks; die "forks version 0.18 required--this is only version $forks::VERSION"
@@ -57,26 +98,41 @@ BEGIN {
 
 	sub _open_env () {
 		### open the base environment ###
-		return $bdb_env = new BerkeleyDB::Env(
+		$bdb_env = new BerkeleyDB::Env(
 			-Home  => ENV_PATH,
 			-Flags => DB_INIT_CDB | DB_CREATE | DB_INIT_MPOOL,
 		) or _croak( "Can't create BerkeleyDB::Env (home=".ENV_PATH."): $BerkeleyDB::Error" );
 		if (USE_BDB_LOCKS) {
-			return $bdb_locksig_env = new BerkeleyDB::Env(
+			$bdb_locksig_env = new BerkeleyDB::Env(
 				-Home  => ENV_PATH_LOCKSIG,
 				-Flags => DB_INIT_CDB | DB_CREATE | DB_INIT_MPOOL,
 			) or _croak( "Can't create BerkeleyDB::Env (home=".ENV_PATH_LOCKSIG."): $BerkeleyDB::Error" );
 		}
+
+		### set base environment file permissions ###
+		my @env_dirs = DEFAULT_ENV_PATHS;
+		my $env_root_regex = quotemeta ENV_ROOT;
+		foreach my $env_dir (@env_dirs) {
+			opendir(ENVDIR, $env_dir);
+			my @env_files = grep(!/^(\.|\.\.)$/, readdir(ENVDIR));
+			closedir(ENVDIR);
+			foreach (@env_files) {
+				my $file = "$env_dir/$_";
+				$file =~ m/^([\/-\@\w_.]+)$/so;	#untaint
+				#TODO: do we need to modify owner and grp to use custom environment settings?
+				chmod BDB_ENV_CHMOD_OCTVAL | 0111, $1;
+			}
+		}
 	}
 
 	sub _close_env () {
 		### close and undefine the base environment ###
-		$bdb_env->close() if defined $bdb_env && UNIVERSAL::isa($_[0], 'BerkeleyDB::Env');
+		eval { $bdb_env->close() };
 		$bdb_env = undef;
 	}
 
 	sub _purge_env (;$) {
-		my @env_dirs = @_ ? @_ : (ENV_PATH, (USE_BDB_LOCKS() ? ENV_PATH_LOCKSIG : ()));
+		my @env_dirs = @_ ? @_ : DEFAULT_ENV_PATHS;
 		foreach my $env_dir (@env_dirs) {
 			opendir(ENVDIR, $env_dir);
 			my @files_to_del = reverse grep(!/^(\.|\.\.)$/, readdir(ENVDIR));
@@ -85,7 +141,8 @@ BEGIN {
 			foreach (@files_to_del) {
 				my $file = "$env_dir/$_";
 				$file =~ m/^([\/-\@\w_.]+)$/so;	#untaint
-				_croak( "Unable to unlink file '$1'. Please manually remove this file." ) unless unlink $1;
+				_croak( "Unable to unlink file '$1'. Please manually remove this file." )
+					unless unlink $1;
 			}
 		}
 	}
@@ -99,7 +156,7 @@ BEGIN {
 	}
 	
 	sub _fork {
-		### safely sync & close databases, close environment ###
+		### safely sync/close databases, close environment ###
 		_untie_support_vars();
 		_close_env();
 		
@@ -119,6 +176,7 @@ BEGIN {
 
 	### create/purge necessary paths to create clean environment ###
 	my @env_dirs = (ENV_PATH, (USE_BDB_LOCKS() ? ENV_PATH_LOCKSIG : ()));
+	my $env_root_regex = quotemeta ENV_ROOT;
 	foreach my $env_dir (@env_dirs) {
 		if (-d $env_dir) {
 			_purge_env($env_dir);
@@ -128,10 +186,14 @@ BEGIN {
 			foreach (split(/\//o, $env_dir)) {
 				$curpath .= $_ eq '' ? '/' : "$_/";
 				unless (-d $curpath) {
-					my $status = mkdir $curpath, 0777;
-					_croak( "Can't create directory ".ENV_ROOT.': '.$! ) unless $status || -d $curpath;
+					my $status = mkdir $curpath, BDB_ENV_CHMOD_OCTVAL | 0111;
+					chown BDB_ENV_CHOWN_ID, BDB_ENV_CHGRP_ID, $curpath
+						unless BDB_ENV_CHOWN_ID == -1 && BDB_ENV_CHGRP_ID == -1;
+					_croak( "Can't create directory ".ENV_ROOT.': '.$! )
+						unless $status || -d $curpath;
 				}
-				chmod 0777, $curpath;
+				chmod BDB_ENV_CHMOD_OCTVAL | 0111, $curpath
+					if $curpath =~ m/^$env_root_regex/o;
 			}
 		}
 	}
@@ -142,11 +204,15 @@ BEGIN {
 }
 
 END {
+	local $@;
 	eval { _untie_support_vars(); };
-	eval { _close_env(); };
+#	eval { _close_env(); };	#disabled: appears to reduce 100% CPU deadlock, main thread
 	#also remove database if no threads connected to any databases (maybe use recno DB to monitor num of threads connected per shared var)?
 }
 
+sub bdb_env { return $bdb_env; }
+sub bdb_locksig_env { return $bdb_locksig_env; }
+
 sub CLONE {	#reopen environment and immediately retie to critical databases
 	_open_env();
 	_tie_support_vars();
@@ -163,7 +229,7 @@ forks::BerkeleyDB - high-performance drop-in replacement for threads
 
 =head1 VERSION
 
-This documentation describes version 0.053.
+This documentation describes version 0.054.
 
 =head1 SYNOPSYS
 
@@ -215,6 +281,33 @@ comparable to native ithreads.
 
 See L<forks> for common usage information.
 
+=head1 Environment Variables
+
+C<forks::BerkeleyDB> supports several environment variables.
+
+=head2 TMPDIR
+
+C<forks::BerkeleyDB> requires a temporary directory to store all BerkeleyDB environment
+and database files.  This variable is controlled by L<File::Spec>, so the default location
+for such files (in the case that TMPDIR is unset) will depend on your platform; e.g.
+File::Spec::Unix checks C<$ENV{TMPDIR}> (unless taint is on) and C</tmp>.
+
+=head2 THREADS_BDB_ENV_CHMOD
+
+Sets the default file and directory permissions of BerkeleyDB environment and database files.
+If unset, will use the Perl default; e.g. current process (thread) L<umask|perlfunc/"umask">
+with defaults for L<mkdir|perlfunc/"mkdir"> and L<open|perlfunc/"open">.
+
+=head2 THREADS_BDB_ENV_CHOWN
+
+Sets the default group owner of BerkeleyDB environment and database files.  If unset,
+will use the Perl default; e.g. current process (thread) effective user.
+
+=head2 THREADS_BDB_ENV_CHGRP
+
+Sets the default group owner of BerkeleyDB environment and database files.  If unset,
+will use the Perl default; e.g. current process (thread) effective group.
+
 =head1 NOTES
 
 If you have forks.pm 0.23 or later installed, all database files created during runtime
@@ -246,53 +339,7 @@ to support chaining, like the following
 
 =head1 TODO
 
-Implement thread joined data using BerkeleyDB.
-
-Determine what additional functions should be migrated to BerkeleyDB backend vs. those that
-should remain as part of the forks package.
-
-Add a high security mode, where all BerkeleyDB data is encrypted using either
-native encryption (preferred, if available) or an external cryptography module
-of the user's choice (i.e. Crypt::* interface module, or something that
-supports a standard interface given an object instance).
-
-Consider porting all shared variable tied class support into package classes,
-instead of depending on BerkeleyDB module parent classes for some methods, to
-insure method behavior consistency no matter which BerkeleyDB.pm version is installed.
-
-Consider merging shared scalars into one or more BDB recno tables, to minimize
-use of environment locks and database files (at the cost of write cursor performance,
-if multiple threads attempting to write to different SVs in same physical table).
-
-Consider rewriting all SV actions to use write cursor (unless complete action is already
-atomic in BDB API) to insure perltie actions are atomic in nature.  Intention is to allow
-use of SV without always requiring a lock (for apps that require highest possible
-concurrency).
-
-Consider implementing "atomic" shared variable classes, which allow all non-iterative
-operations to be atomic without locks.  This would require overload of all math and
-string operators.  Hopefully this will be enabled with an attribute, such
-as 'sharedatomic'.  I don't believe this can be achieved with perltie, so only non-blessed
-primitives would be allowed for scalars.
-
-May need to enable DB_ENV->failchk when shared var process detects that a thread
-has unexpectedly exited.  If return value is DB_RUNRECOVERY, then we likely need
-to terminate the entire application (as the shared bdb environment is no longer
-guaranteed to be stable.
-
-Consider using bdb txn subsystem environment for locking and signaling. Theoretically,
-this should require: 1 recno for locks (idx=sid, value=tid holding lock), 1 recno for waiting
-(idx=sid, value=[list if tid waiting]), and N queue for signaling (1 per thread; thread block 
-on own queue to wimulate waiting; push from other source acts as signal).  Txn would be
-used on locks and waiting recno databases (locking individual elements with cursors). Deadlock
-detection could be enabled using BDB deadlock detection engine.  Would need hooks into deadlock
-detection forks.pm interface.
-
-=head1 CAVIATS
-
-It appears that BerkeleyDB libdb 4.4.x environments are not fully thread-safe
-with BerkeleyDB CDB mode on some platforms.  Thus, it is highly recommended you
-use libdb 4.3.x and earlier, or 4.5.x and later.
+See the TODO file in the distribution.
 
 =head1 AUTHOR
 
@@ -300,7 +347,7 @@ Eric Rybski <rybskej@yahoo.com>.
 
 =head1 COPYRIGHT
 
-Copyright (c) 2006-2007 Eric Rybski <rybskej@yahoo.com>.
+Copyright (c) 2006-2008 Eric Rybski <rybskej@yahoo.com>.
 All rights reserved.  This program is free software; you can redistribute it
 and/or modify it under the same terms as Perl itself.
 
